##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Set's up the reactor used by zrs
"""

import atexit, logging, os, threading

import twisted.internet
import twisted.python.log

logger = logging.getLogger(__name__)

def log_twisted(data):
    message = data['message']
    message = '\n'.join([str(m) for m in message])
    if data['isError']:
        logger.error(message)
    else:
        logger.info(message)

twisted.python.log.startLoggingWithObserver(log_twisted, setStdout=False)

_shutdown_called = False
_started = False

def _run(reactor):
    global _shutdown_called
    try:
        reactor.run(False)
    except:
        logger.exception("The reactor errored")

    if not _shutdown_called:
        logger.critical("The twisted reactor quit unexpectedly")
        _shutdown_called = True
        #logging.shutdown()
        # Force the process to exit. I wish there was a less violent way.
        #os._exit(os.EX_SOFTWARE)

def reactor():
    global _started
    if not _started:
        _started = True
        from twisted.internet import reactor
        thread = threading.Thread(target=_run,
                                  args=(twisted.internet.reactor, ))
        thread.setDaemon(True)
        thread.start()
        atexit.register(shutdown)

    return twisted.internet.reactor

def _shutdown(event):
    twisted.internet.reactor.stop()
    event.set()

def shutdown():
    global _shutdown_called
    if not _started or _shutdown_called:
        return
    
    _shutdown_called = True
    event = threading.Event()
    twisted.internet.reactor.callFromThread(_shutdown, event)
    event.wait(60)




##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import cPickle
import logging
import os
import sys
import threading
import time
import twisted.internet.interfaces
import twisted.internet.protocol
import zc.zrs.reactor
import zc.zrs.sizedmessage
import ZODB.BaseStorage
import ZODB.blob
import ZODB.FileStorage
import ZODB.FileStorage.format
import ZODB.interfaces
import ZODB.TimeStamp
import ZODB.utils
import zope.interface

if sys.version_info > (2, 5):
    from hashlib import md5
else:
    from md5 import md5

if not hasattr(ZODB.blob.BlobStorage, 'restoreBlob'):
    import zc.zrs.restoreblob

logger = logging.getLogger(__name__)

class Base(object):

    def __init__(self, storage, addr, reactor):
        if reactor is None:
            reactor = zc.zrs.reactor.reactor()

        self._reactor = reactor

        self._storage = storage

        try:
            fn = storage.getExtensionMethods
        except AttributeError:
            pass # no extension methods
        else:
            self.getExtensionMethods = storage.getExtensionMethods
            for name in fn():
                assert not hasattr(self, name)
                setattr(self, name, getattr(storage, name))

        self._addr = addr

    def __len__(self):
        return len(self._storage)

class Primary(Base):

    def __init__(self, storage, addr, reactor=None):
        Base.__init__(self, storage, addr, reactor)

        if ZODB.interfaces.IBlobStorage.providedBy(storage):
            zope.interface.directlyProvides(self, ZODB.interfaces.IBlobStorage)
            for name in ('storeBlob', 'loadBlob', 'temporaryDirectory',
                         'restoreBlob', 'openCommittedBlobFile', 'fshelper'):
                setattr(self, name, getattr(storage, name))

        if not isinstance(storage, (ZODB.blob.BlobStorage,
                                    ZODB.FileStorage.FileStorage)
                          ):
            raise ValueError("Invalid storage", storage)

        self._changed = threading.Condition()

        # required methods
        for name in (
            'getName', 'getSize', 'history', 'isReadOnly', 'lastTransaction',
            'load', 'loadBefore', 'loadSerial', 'new_oid', 'pack',
            'registerDB', 'sortKey', 'store', 'tpc_abort', 'tpc_begin',
            'tpc_vote',
            ):
            setattr(self, name, getattr(storage, name))

        # Optional methods:
        for name in (
            'restore', 'iterator', 'cleanup', 'loadEx', 'getSerial',
            'supportsTransactionalUndo',
            'tpc_transaction', 'getTid', 'lastInvalidations',
            'supportsUndo', 'undoLog', 'undoInfo', 'undo',
            'supportsVersions', 'abortVersion', 'commitVersion',
            'versionEmpty', 'modifiedInVersion', 'versions',
            'record_iternext', 'deleteObject',
            'checkCurrentSerialInTransaction',
            ):
            if hasattr(storage, name):
                setattr(self, name, getattr(storage, name))

        self._factory = PrimaryFactory(storage, self._changed)
        logger.info("Opening %s %s", self.getName(), addr)
        self._reactor.callFromThread(self.cfr_listen)

    # StorageServer accesses _transaction directly. :(
    @property
    def _transaction(self):
        return self._storage._transaction

    _listener = None
    def cfr_listen(self):
        if isinstance(self._addr, basestring):
            self._listener = self._reactor.listenUNIX(self._addr, self._factory)
        else:
            interface, port = self._addr
            self._listener = self._reactor.listenTCP(
                port, self._factory, interface=interface)

    def cfr_stop_listening(self):
        if self._listener is not None:
            self._listener.stopListening()

    def tpc_finish(self, *args):
        self._storage.tpc_finish(*args)
        self._changed.acquire()
        self._changed.notifyAll()
        self._changed.release()

    def close(self):
        logger.info('Closing %s %s', self.getName(), self._addr)
        self._reactor.callFromThread(self.cfr_stop_listening)

        # Close the storage first to prevent more writes and to
        # give the secondaries more time to catch up.
        self._storage.close()
        self._factory.close()

class PrimaryFactory(twisted.internet.protocol.Factory):

    def __init__(self, storage, changed):
        self.storage = storage
        self.changed = changed
        self.instances = []
        self.threads = ThreadCounter()

    def close(self):
        for instance in list(self.instances):
            instance.close()
        self.threads.wait(60)

class PrimaryProtocol(twisted.internet.protocol.Protocol):

    __protocol = None
    __start = None
    __producer = None

    def connectionMade(self):           # cfr
        self.__stream = zc.zrs.sizedmessage.Stream(self.messageReceived, 8)
        self.__peer = str(self.transport.getPeer()) + ': '
        self.factory.instances.append(self)
        self.info("Connected")

    def connectionLost(self, reason):   # cfr
        self.info("Disconnected %r", reason)
        if self.__producer is not None:
            self.__producer.stopProducing()
        self.factory.instances.remove(self)

    def close(self):
        if self.__producer is not None:
            self.__producer.close()
        else:
            self.transport.reactor.callFromThread(
                self.transport.loseConnection)
        self.info('Closed')

    def _stop(self):
        # for tests
        if self.__producer is not None:
            self.__producer._stop()
        else:
            self.transport.reactor.callFromThread(
                self.transport.loseConnection)


    def error(self, message, *args):
        logger.error(self.__peer + message, *args)
        self.close()

    def info(self, message, *args):
        logger.info(self.__peer + message, *args)

    def dataReceived(self, data):       # cfr
        try:
            self.__stream(data)
        except zc.zrs.sizedmessage.LimitExceeded, v:
            self.error('message too large: '+str(v))

    def messageReceived(self, data):    # cfr
        if self.__protocol is None:
            if data == 'zrs2.0':
                if ZODB.interfaces.IBlobStorage.providedBy(
                    self.factory.storage):
                    return self.error("Invalid protocol %r. Require >= 2.1",
                                      data)
            elif data != 'zrs2.1':
                return self.error("Invalid protocol %r", data)
            self.__protocol = data
        else:
            if self.__start is not None:
                if not data:
                    logger.debug(self.__peer + "keep-alive")
                    return # ignore empty messages
                return self.error("Too many messages")
            self.__start = data
            if len(data) != 8:
                return self.error("Invalid transaction id, %r", data)

            self.info("start %r (%s)", data, ZODB.TimeStamp.TimeStamp(data))
            self.__producer = PrimaryProducer(
                (self.factory.storage, self.factory.changed, self.__start),
                self.transport, self.__peer,
                self.factory.threads.run)

PrimaryFactory.protocol = PrimaryProtocol


class PrimaryProducer:

    zope.interface.implements(twisted.internet.interfaces.IPushProducer)

    # stopped indicates that we should stop sending output to a client
    stopped = False

    # closed means that we're closed.  This is mainly to deal with a
    # possible race at startup.  When we close a producer, we continue
    # sending data to the client as long as data are available.
    closed = False

    def __init__(self, iterator_args, transport, peer,
                 run=lambda f, *args: f(*args)):
        self.iterator_scan_control = ScanControl()
        self.storage = iterator_args[0]
        self.iterator = None
        self.iterator_args = iterator_args + (self.iterator_scan_control,)
        self.start_tid = iterator_args[2]
        self.transport = transport
        self.peer = peer
        transport.registerProducer(self, True)
        self.callFromThread = transport.reactor.callFromThread
        self.consumer_event = threading.Event()
        self.consumer_event.set()
        thread = threading.Thread(target=run, args=(self.run, ),
                                  name='Producer(%s)' % peer)
        thread.setDaemon(True)
        thread.start()
        self.thread = thread

    def pauseProducing(self):
        logger.debug(self.peer+" pausing")
        self.consumer_event.clear()

    def resumeProducing(self):
        logger.debug(self.peer+" resuming")
        self.consumer_event.set()


    def close(self):
        # We use the closed flag to handle a race condition in
        # iterator setup.  We set the close flag before checking for
        # the iterator.  If we find the iterator, we tell it to stop.
        # If not, then when the run method below finishes creating the
        # iterator, it will find the close flag and not use the iterator.
        self.closed = True

        iterator = self.iterator
        if iterator is not None:
            iterator.catch_up_then_stop()

    def _stop(self):
        # for tests
        iterator = self.iterator
        if iterator is not None:
            iterator.stop()

    def cfr_close(self):
        if not self.stopped:
            self.transport.unregisterProducer()
            self.stopProducing()
            self.transport.loseConnection()


    def stopProducing(self): # cfr
        self.stopped = True
        iterator = self.iterator
        if iterator is not None:
            iterator.stop()
        else:
            self.iterator_scan_control.not_stopped = False

        self.consumer_event.set() # unblock wait calls

    def cfr_write(self, data):
        if not self.stopped:
            self.transport.writeSequence(data)

    def write(self, data):
        data = zc.zrs.sizedmessage.marshals(data)
        self.md5.update(data[1])
        self.consumer_event.wait()
        self.callFromThread(self.cfr_write, data)


    def run(self):
        try:
            self.iterator = FileStorageIterator(*self.iterator_args)
        except:
            logger.exception(self.peer)
            self.iterator = None
            self.callFromThread(self.cfr_close)
            return

        if self.closed:
            self.callFromThread(self.cfr_close)
            return

        if self.stopped:
            # make sure our iterator gets stopped, as stopProducing might
            # have been called while we were creating the iterator.
            self.iterator.stop()

        pickler = cPickle.Pickler(1)
        pickler.fast = 1
        self.md5 = md5(self.start_tid)

        blob_block_size = 1 << 16

        try:
            for trans in self.iterator:
                self.write(
                    pickler.dump(('T', (trans.tid, trans.status, trans.user,
                                        trans.description, trans._extension)),
                                 1)
                    )
                for record in trans:
                    if record.data and is_blob_record(record.data):
                        try:
                            fname = self.storage.loadBlob(
                                record.oid, record.tid)
                            f = open(fname, 'rb')
                        except (IOError, ZODB.POSException.POSKeyError):
                            pass
                        else:
                            f.seek(0, 2)
                            blob_size = f.tell()
                            blocks, r = divmod(blob_size, blob_block_size)
                            if r:
                                blocks += 1

                            self.write(
                                pickler.dump(
                                    ('B',
                                     (record.oid, record.tid, record.version,
                                      record.data_txn, long(blocks))),
                                    1)
                                )
                            self.write(record.data or '')
                            f.seek(0)
                            while blocks > 0:
                                data = f.read(blob_block_size)
                                if not data:
                                    raise AssertionError("Too much blob data")
                                blocks -= 1
                                self.write(data)

                            f.close()
                            continue

                    self.write(
                        pickler.dump(('S',
                                       (record.oid, record.tid, record.version,
                                        record.data_txn)),
                                     1)
                        )
                    self.write(record.data or '')

                self.write(pickler.dump(('C', (self.md5.digest(), )), 1))

        except:
            logger.exception(self.peer)

        self.iterator = None
        self.callFromThread(self.cfr_close)


class TidTooHigh(Exception):
    """The last tid for an iterator is higher than any tids in a file.
    """

class ScanControl:

    not_stopped = True

class FileStorageIterator(ZODB.FileStorage.format.FileStorageFormatter):

    _file_size = 1 << 64 # To make base class check happy.

    def __init__(self, fs, condition=None, start=ZODB.utils.z64,
                 scan_control=None):
        self._ltid = start
        self._fs = fs
        self._stop = False
        if scan_control is None:
            scan_control = ScanControl()
        self._scan_control = scan_control
        self._open()
        if condition is None:
            condition = threading.Condition()
        self._condition = condition
        self._catch_up_then_stop = False

    def _open(self):
        self._old_file = self._fs._file
        try:
            file = open(self._fs._file_name, 'rb', 0)
        except IOError, v:
            if os.path.exists(self._fs._file_name):
                raise

            # The file is gone.  It must have been removed -- probably
            # in a test.  We won't die here because we're probably
            # never going to be used.  If we are ever used

            def _next():
                raise v
            self._next = _next
            return

        self._file = file
        self._pos = 4L
        ltid = self._ltid
        if ltid > ZODB.utils.z64:
            # We aren't starting at the beginning.  We need to find the
            # first transaction after _ltid. We can either search from the
            # beginning, or from the end.
            file.seek(4)
            tid = file.read(8)
            if len(tid) < 8:
                raise TidTooHigh(repr(ltid))

            t1 = ZODB.TimeStamp.TimeStamp(tid).timeTime()
            tid = self._fs.lastTransaction()
            t2 = ZODB.TimeStamp.TimeStamp(tid).timeTime()
            t = ZODB.TimeStamp.TimeStamp(self._ltid).timeTime()

            if (t - t1) < (t2 - t1)/2:
                # search from beginning:
                self._scan_forward(4, ltid)
            else:
                # search from end
                pos = self._fs._pos - 8
                if pos < 27:
                    # strangely small position
                    return self._scan_forward(4, ltid)

                file.seek(pos)
                tlen = ZODB.utils.u64(file.read(8))
                pos -= tlen
                if pos <= 4:
                    # strangely small position
                    return self._scan_forward(4, ltid)

                file.seek(pos)
                h = self._read_txn_header(pos)
                if h.tid <= ltid:
                    self._scan_forward(pos, ltid)
                else:
                    self._scan_backward(pos, ltid)

    def _scan_forward(self, pos, ltid):
        file = self._file
        while self._scan_control.not_stopped:
            # Read the transaction record
            try:
                h = self._read_txn_header(pos)
            except ZODB.FileStorage.format.CorruptedDataError, err:
                # If buf is empty, we've reached EOF.
                if not err.buf:
                    # end of file.
                    self._pos = pos
                    raise TidTooHigh(repr(ltid))
                raise

            if h.status == "c":
                raise TidTooHigh(repr(ltid))

            if h.tid > ltid:
                # This is the one we want to read next
                self._pos = pos
                return

            pos += h.tlen + 8
            if h.tid == ltid:
                # We just read the one we want to skip past
                self._pos = pos
                return

    def _scan_backward(self, pos, ltid):
        file = self._file
        seek = file.seek
        read = file.read
        while self._scan_control.not_stopped:
            pos -= 8
            seek(pos)
            tlen = ZODB.utils.u64(read(8))
            pos -= tlen
            h = self._read_txn_header(pos)
            if h.tid <= ltid:
                self._pos = pos + tlen + 8
                return

    def __iter__(self):
        return self

    def stop(self):
        self._condition.acquire()
        self._stop = True
        self._condition.notifyAll()
        self._condition.release()

    def catch_up_then_stop(self):
        self._condition.acquire()
        self._catch_up_then_stop = True
        self._condition.notifyAll()
        self._condition.release()

    def next(self):
        self._condition.acquire()
        try:
            while 1:
                if self._stop:
                    raise StopIteration
                r = self._next()
                if r is not None:
                    return r
                if self._catch_up_then_stop:
                    raise StopIteration
                self._condition.wait()
        finally:
            self._condition.release()

    def _next(self):

        if self._old_file is not self._fs._file:
            # Our file-storage must have been packed.  We need to
            # reopen the file:
            self._open()

        file = self._file
        pos = self._pos

        while 1:
            # Read the transaction record
            try:
                h = self._read_txn_header(pos)
            except ZODB.FileStorage.format.CorruptedDataError, err:
                # If buf is empty, we've reached EOF.
                if not err.buf:
                    return None
                raise

            if h.tid <= self._ltid:
                logger.warning("%s time-stamp reduction at %s",
                               self._file.name, pos)

            if h.status == "c":
                # Assume we've hit the last, in-progress transaction
                # Wait until there is more data.
                return None

            if pos + h.tlen + 8 > self._file_size:
                # Hm, the data were truncated or the checkpoint flag wasn't
                # cleared.  They may also be corrupted,
                # in which case, we don't want to totally lose the data.
                logger.critical("%s truncated, possibly due to"
                                " damaged records at %s", self._file.name, pos)
                raise ZODB.FileStorage.format.CorruptedDataError(None, '', pos)

            if h.status not in " up":
                logger.warning('%s has invalid status,'
                               ' %s, at %s', self._file.name, h.status, pos)

            if h.tlen < h.headerlen():
                logger.critical("%s has invalid transaction header at %s",
                                self._file.name, pos)
                raise ZODB.FileStorage.format.CorruptedDataError(None, '', pos)

            tpos = pos
            tend = tpos + h.tlen

            # Check the (intentionally redundant) transaction length
            self._file.seek(tend)
            rtl = ZODB.utils.u64(self._file.read(8))
            if rtl != h.tlen:
                logger.warning("%s redundant transaction length check"
                               " failed at %s", self._file.name, tend)
                break

            self._pos = tend + 8
            self._ltid = h.tid

            if h.status == "u":
                # Undone trans.  It's unlikely to see these anymore
                continue

            pos += h.headerlen()
            if h.elen:
                e = cPickle.loads(h.ext)
            else:
                e = {}

            result = RecordIterator(
                h.tid, h.status, h.user, h.descr,
                e, pos, tend, self._file, tpos)

            return result

    def notify(self):
        self._condition.acquire()
        self._condition.notifyAll()
        self._condition.release()

class RecordIterator(ZODB.FileStorage.format.FileStorageFormatter):
    """Iterate over the transactions in a FileStorage file."""

    def __init__(self, tid, status, user, desc, ext, pos, tend, file, tpos):
        self.tid = tid
        self.status = status
        self.user = user
        self.description = desc
        self._extension = ext
        self._pos = pos
        self._tend = tend
        self._file = file
        self._tpos = tpos

    def __iter__(self):
        return self

    def next(self):
        pos = self._pos
        if pos < self._tend:
            # Read the data records for this transaction
            h = self._read_data_header(pos)
            dlen = h.recordlen()

            if pos + dlen > self._tend or h.tloc != self._tpos:
                logger.critical("%s data record exceeds transaction"
                                " record at %s", file.name, pos)
                raise ZODB.FileStorage.format.CorruptedDataError(
                    h.oid, '', pos)

            self._pos = pos + dlen
            prev_txn = None
            if h.plen:
                data = self._file.read(h.plen)
            else:
                if h.back == 0:
                    # If the backpointer is 0, then this transaction
                    # undoes the object creation.  It either aborts
                    # the version that created the object or undid the
                    # transaction that created it.  Return None
                    # instead of a pickle to indicate this.
                    data = None
                else:
                    data, tid = self._loadBackTxn(h.oid, h.back, False)
                    # Caution:  :ooks like this only goes one link back.
                    # Should it go to the original data like BDBFullStorage?
                    prev_txn = self.getTxnFromData(h.oid, h.back)

            return Record(h.oid, h.tid, '', data, prev_txn, pos)

        raise StopIteration

class Record:
    """An abstract database record."""
    def __init__(self, oid, tid, version, data, prev, pos):
        self.oid = oid
        self.tid = tid
        self.version = version
        self.data = data
        self.data_txn = prev
        self.pos = pos

class ThreadCounter:
    """Keep track of running threads

    We use the run method to run the threads and the wait method to
    wait until they're all done:
    """

    def __init__(self):
        self.count = 0
        self.condition = threading.Condition()

    def run(self, func, *args):
        self.condition.acquire()
        self.count += 1
        self.condition.release()
        try:
            func(*args)
        finally:
            self.condition.acquire()
            self.count -= 1
            self.condition.notifyAll()
            self.condition.release()

    def wait(self, timeout):
        deadline = time.time()+timeout
        self.condition.acquire()
        while self.count > 0:
            w = deadline-time.time()
            if w <= 0:
                break
            self.condition.wait(w)
        self.condition.release()

def is_blob_record(record):
    try:
        return cPickle.loads(record) is ZODB.blob.Blob
    except (MemoryError, KeyboardInterrupt, SystemExit):
        raise
    except Exception:
        return False



##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import struct

def marshal(message):
    return struct.pack(">I", len(message)) + message

def marshals(message):
    return struct.pack(">I", len(message)), message

class LimitExceeded(Exception):
    """Too much data
    """

class Stream:

    def __init__(self, callback, limit = 1 << 32):
        self._callback = callback
        self.limit = limit
        self.data = []
        self.size = 0
        self.length = 0

    def __call__(self, data):
        self.data.append(data)
        self.size += len(data)
        while 1:
            if self.length:
                if self.size < self.length:
                    return

                data = ''.join(self.data)
                result = data[:self.length]
                data = data[self.length:]
                self.data = [data]
                self.size = len(data)
                self.length = 0
                self._callback(result)

            if self.size < 4:
                return

            if len(self.data[0]) < 4:
                self.data = [''.join(self.data)]

            self.length, = struct.unpack(">I", self.data[0][:4])
            self.data[0] = self.data[0][4:]
            self.size -= 4
            if self.length == 0:
                self._callback('')
            elif self.length > self.limit:
                raise LimitExceeded(self.limit, long(self.length))




##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ZooKeeper support
"""
import zc.zrs.primary
import zc.zrs.secondary

class Primary(zc.zrs.primary.Primary):

    def __init__(self, storage, addr, connection_string, path):
        self.__zk = zc.zk.ZooKeeper(connection_string)
        self.__path = path
        super(Primary, self).__init__(storage, addr)

    def cfr_listen(self):
        super(Primary, self).cfr_listen()
        try:
            addr = self._listener.getHost()
            self.__zk.register_server(self.__path, (self._addr[0], addr.port))
        except:
            import traceback; traceback.print_exc()
            raise

    def close(self):
        self.__zk.close()
        super(Primary, self).close()

def first_addr(addrs):
    host, port = iter(addrs).next().split(':')
    return host, int(port)

class Secondary(zc.zrs.secondary.Secondary):

    def __init__(self, storage, connection_string, path, *args, **kw):
        self.__zk = zc.zk.ZooKeeper(connection_string)
        self.__addrs = self.__zk.children(path)
        super(Secondary, self).__init__(
            storage, first_addr(self.__addrs) if self.__addrs else None,
            *args, **kw)

        @self.__addrs
        def update(addrs):
            if addrs:
                self.setReplicationAddress(first_addr(addrs))

    def close(self):
        self.__zk.close()
        super(Secondary, self).close()



# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
# See LICENSE for details.

"""Loopback-transport support

This is derived from and will eventually feed back to
twisted.protocols.loopback.
"""


# system imports
from zope.interface import implements

# Twisted Imports
from twisted.internet import interfaces, protocol, main, defer
from twisted.python import failure
from twisted.internet.interfaces import IAddress

class _LoopbackQueue(object):
    """
    Trivial wrapper around a list to give it an interface like a queue, which
    the addition of also sending notifications by way of a Deferred whenever
    the list has something added to it.
    """

    _notificationDeferred = None
    disconnect = False

    def __init__(self):
        self._queue = []


    def put(self, v):
        self._queue.append(v)
        if self._notificationDeferred is not None:
            d, self._notificationDeferred = self._notificationDeferred, None
            d.callback(None)


    def __nonzero__(self):
        return bool(self._queue)


    def get(self):
        return self._queue.pop(0)



class _LoopbackAddress(object):
    implements(IAddress)


class _LoopbackTransport(object):
    implements(interfaces.ITransport, interfaces.IConsumer)

    disconnecting = False
    producer = None

    # ITransport
    def __init__(self, q, proto):
        self.q = q
        self.proto = proto

    def write(self, bytes):
        self.q.put(bytes)

    def writeSequence(self, iovec):
        self.q.put(''.join(iovec))

    def loseConnection(self):
        self.q.disconnect = True
        self.q.put('')

    def getPeer(self):
        return _LoopbackAddress()

    def getHost(self):
        return _LoopbackAddress()

    # IConsumer
    def registerProducer(self, producer, streaming):
        assert self.producer is None
        self.producer = producer
        self.streamingProducer = streaming
        self._pollProducer()

    def unregisterProducer(self):
        assert self.producer is not None
        self.producer = None

    def _pollProducer(self):
        if self.producer is not None and not self.streamingProducer:
            self.producer.resumeProducing()

    def connectionLost(self, reason):
        if self.producer is not None:
            self.producer.stopProducing()
            self.unregisterProducer()
        self.proto.connectionLost(reason)

class _ClientLoopbackTransport(_LoopbackTransport):

    connector = None
    def connectionLost(self, reason):
        self.proto.connectionLost(reason)
        if self.connector is not None:
            self.connector.connectionLost(reason)

def loopbackAsync(server, client, connector):
    serverToClient = _LoopbackTransport(_LoopbackQueue(), server)
    serverToClient.reactor = connector.reactor

    clientToServer = _ClientLoopbackTransport(_LoopbackQueue(), client)
    clientToServer.reactor = connector.reactor
    clientToServer.connector = connector

    server.makeConnection(serverToClient)
    client.makeConnection(clientToServer)

    _loopbackAsyncBody(serverToClient, clientToServer)



def _loopbackAsyncBody(serverToClient, clientToServer):

    def pump(source, q, target):
        sent = False
        while q:
            sent = True
            bytes = q.get()
            if bytes:
                target.dataReceived(bytes)

        # A write buffer has now been emptied.  Give any producer on that side
        # an opportunity to produce more data.
        source.transport._pollProducer()

        return sent

    server = serverToClient.proto
    client = clientToServer.proto

    while 1:
        disconnect = clientSent = serverSent = False

        # Deliver the data which has been written.
        serverSent = pump(server, serverToClient.q, client)
        clientSent = pump(client, clientToServer.q, server)

        if not clientSent and not serverSent:
            # Neither side wrote any data.  Wait for some new data to be added
            # before trying to do anything further.
            d = defer.Deferred()
            clientToServer.q._notificationDeferred = d
            serverToClient.q._notificationDeferred = d
            d.addCallback(_loopbackAsyncContinue,
                          serverToClient, clientToServer)
            break

        if serverToClient.q.disconnect:
            # The server wants to drop the connection.  Flush any remaining
            # data it has.
            disconnect = True
            pump(server, serverToClient.q, client)
        elif clientToServer.q.disconnect:
            # The client wants to drop the connection.  Flush any remaining
            # data it has.
            disconnect = True
            pump(client, clientToServer.q, server)

        if disconnect:
            # Someone wanted to disconnect, so okay, the connection is gone.
            serverToClient.connectionLost(
                failure.Failure(main.CONNECTION_DONE))
            clientToServer.connectionLost(
                failure.Failure(main.CONNECTION_DONE))
            break


def _loopbackAsyncContinue(ignored, serverToClient, clientToServer):
    # Clear the Deferred from each message queue, since it has already fired
    # and cannot be used again.
    clientToServer.q._notificationDeferred = None
    serverToClient.q._notificationDeferred = None

    # Push some more bytes around.
    _loopbackAsyncBody(serverToClient, clientToServer)



##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import cPickle
import logging
import md5
import os
import tempfile
import threading
import twisted.internet.protocol
import zc.zrs.primary
import zc.zrs.reactor
import zc.zrs.sizedmessage
import ZODB.blob
import ZODB.interfaces
import ZODB.POSException
import zope.interface

logger = logging.getLogger(__name__)


class SecondaryProtocol(twisted.internet.protocol.Protocol):

    _zrs_transaction = None

    keep_alive_delayed_call = None

    logger = logger

    def connectionMade(self):
        self.__stream = zc.zrs.sizedmessage.Stream(self.messageReceived)
        self.__peer = str(self.transport.getPeer()) + ': '
        self.factory.instance = self
        self.transport.write(zc.zrs.sizedmessage.marshal(
            self.factory.zrs_proto))
        tid = self.factory.storage.lastTransaction()
        self._replication_stream_md5 = md5.new(tid)
        self.transport.write(zc.zrs.sizedmessage.marshal(tid))
        self.info("Connected")
        if self.factory.keep_alive_delay > 0:
            self.keep_alive()

    def connectionLost(self, reason):
        if (self.keep_alive_delayed_call is not None
            and self.keep_alive_delayed_call.active()):
            self.keep_alive_delayed_call.cancel()

        self.factory.instance = None
        if self._zrs_transaction is not None:
            self.factory.storage.tpc_abort(self._zrs_transaction)
            self._zrs_transaction = None
        self.info("Disconnected %r", reason)

    def error(self, message, *args, **kw):
        self.logger.critical(self.__peer + message, *args, **kw)
        self.factory.connector.disconnect()

    def info(self, message, *args):
        self.logger.info(self.__peer + message, *args)

    def keep_alive(self):
        if self.keep_alive_delayed_call is not None:
            self.transport.write("\0\0\0\0")
        self.keep_alive_delayed_call = self.factory.reactor.callLater(
            self.factory.keep_alive_delay,
            self.keep_alive,
            )

    def dataReceived(self, data):
        try:
            self.__stream(data)
        except:
            self.error("Input data error", exc_info=True)


    __blob_file_blocks = None
    __blob_file_handle = None
    __blob_file_name = None
    __blob_record = None
    __record = None
    def messageReceived(self, message):
        if self.__record:
            # store or store blob data record
            oid, serial, version, data_txn = self.__record
            self.__record = None
            data = message or None
            key = serial, version
            oids = self.__inval.get(key)
            if oids is None:
                oids = self.__inval[key] = {}
            oids[oid] = 1

            if self.__blob_file_blocks:
                # We have to collect blob data
                self.__blob_record = oid, serial, data, version, data_txn

                (self.__blob_file_handle, self.__blob_file_name
                 ) = tempfile.mkstemp('blob', 'secondary',
                                      self.factory.storage.temporaryDirectory()
                                      )
            else:
                self.factory.storage.restore(
                    oid, serial, data, version, data_txn,
                    self._zrs_transaction)

        elif self.__blob_record:
            os.write(self.__blob_file_handle, message)
            self.__blob_file_blocks -= 1
            if self.__blob_file_blocks == 0:
                # We're done collecting data, we can write the data:
                os.close(self.__blob_file_handle)
                oid, serial, data, version, data_txn = self.__blob_record
                self.__blob_record = None
                self.factory.storage.restoreBlob(
                    oid, serial, data, self.__blob_file_name, data_txn,
                    self._zrs_transaction)

        else:
            # Ordinary message
            message_type, data = cPickle.loads(message)
            if message_type == 'T':
                assert self._zrs_transaction is None
                assert self.__record is None
                transaction = Transaction(*data)
                self.__inval = {}
                self.factory.storage.tpc_begin(
                    transaction, transaction.id, transaction.status)
                self._zrs_transaction = transaction
            elif message_type == 'S':
                self.__record = data
            elif message_type == 'B':
                self.__record = data[:-1]
                self.__blob_file_blocks = data[-1]
            elif message_type == 'C':
                self._check_replication_stream_checksum(data)
                assert self._zrs_transaction is not None
                assert self.__record is None
                self.factory.storage.tpc_vote(self._zrs_transaction)

                def invalidate(tid):
                    if self.factory.db is not None:
                        for (tid, version), oids in self.__inval.items():
                            self.factory.db.invalidate(
                                tid, oids, version=version)

                self.factory.storage.tpc_finish(
                    self._zrs_transaction, invalidate)
                self._zrs_transaction = None
            else:
                raise ValueError("Invalid message type, %r" % message_type)

        self._replication_stream_md5.update(message)

    def _check_replication_stream_checksum(self, data):
        if self.factory.check_checksums:
            checksum = data[0]
            if checksum != self._replication_stream_md5.digest():
                raise AssertionError(
                    "Bad checksum", checksum,
                    self._replication_stream_md5.digest())

class SecondaryFactory(twisted.internet.protocol.ClientFactory):

    db = None
    connector = None
    closed = False
    protocol = SecondaryProtocol

    # We'll keep track of the connected instance, if any mainly
    # for the convenience of some tests that want to force disconnects to
    # stress the secondaries.
    instance = None

    def __init__(self, reactor, storage, reconnect_delay, check_checksums,
                 zrs_proto, keep_alive_delay, secondary):
        self.reactor = reactor
        self.storage = storage
        self.reconnect_delay = reconnect_delay
        self.check_checksums = check_checksums
        self.zrs_proto = zrs_proto
        self.keep_alive_delay = keep_alive_delay
        self.secondary = secondary

    def close(self, callback):
        self.closed = True
        if self.connector is not None:
            self.connector.disconnect()
        callback()

    def startedConnecting(self, connector):
        if self.closed:
            connector.disconnect()
        else:
            self.connector = connector

    def clientConnectionFailed(self, connector, reason):
        self.connector = None
        if not self.closed:
            self.reactor.callLater(self.reconnect_delay, self.connect)

    def clientConnectionLost(self, connector, reason):
        self.connector = None
        if not self.closed:
            self.reactor.callLater(self.reconnect_delay, self.connect)

    def connect(self):
        addr = self.secondary._addr
        reactor = self.reactor
        if isinstance(addr, basestring):
            reactor.callFromThread(reactor.connectUNIX, addr, self)
        else:
            host, port = addr
            reactor.callFromThread(reactor.connectTCP, host, port, self)


class Secondary(zc.zrs.primary.Base):

    factoryClass = SecondaryFactory
    logger = logger

    def __init__(self, storage, addr, reactor=None, reconnect_delay=60,
                 check_checksums=True, keep_alive_delay=0):
        zc.zrs.primary.Base.__init__(self, storage, addr, reactor)

        reactor = self._reactor

        zrs_proto = self.copyMethods(storage)

        self._factory = self.factoryClass(
            reactor, storage, reconnect_delay,
            check_checksums, zrs_proto, keep_alive_delay, self)
        self.logger.info("Opening %s %s", self.getName(), addr)

        if addr:
            self._factory.connect()

    def setReplicationAddress(self, addr):
        old = self._addr
        self._addr = addr
        if not old:
            self._factory.connect()

    def copyMethods(self, storage):
        if (ZODB.interfaces.IBlobStorage.providedBy(storage)
            and hasattr(storage, 'restoreBlob')
            ):
            zope.interface.directlyProvides(self, ZODB.interfaces.IBlobStorage)
            for name in ('loadBlob', 'openCommittedBlobFile',
                         'temporaryDirectory', 'restoreBlob'):
                setattr(self, name, getattr(storage, name))
            zrs_proto = 'zrs2.1'
        else:
            zrs_proto = 'zrs2.0'

        # required methods
        for name in (
            'getName', 'getSize', 'history', 'lastTransaction',
            'load', 'loadBefore', 'loadSerial', 'pack',
            'sortKey',
            ):
            setattr(self, name, getattr(storage, name))

        # Optional methods:
        for name in (
            'iterator', 'cleanup', 'loadEx', 'getSerial',
            'supportsTransactionalUndo',
            'tpc_transaction', 'getTid', 'lastInvalidations',
            'supportsUndo', 'undoLog', 'undoInfo',
            'supportsVersions',
            'versionEmpty', 'modifiedInVersion', 'versions',
            'record_iternext', 'checkCurrentSerialInTransaction',
            ):
            if hasattr(storage, name):
                setattr(self, name, getattr(storage, name))

        return zrs_proto

    def isReadOnly(self):
        return True

    def write_method(*args, **kw):
        raise ZODB.POSException.ReadOnlyError()
    new_oid = tpc_begin = undo = write_method

    def registerDB(self, db, limit=None):
        self._factory.db = db

    def close(self):
        self.logger.info('Closing %s %s', self.getName(), self._addr)
        event = threading.Event()
        self._reactor.callFromThread(self._factory.close, event.set)
        event.wait()
        self._storage.close()


class Transaction:

    def __init__(self, tid, status, user, description, extension):
        self.id = tid
        self.status = status
        self.user = user
        self.description = description
        self._extension = extension




##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging
import sys
import time
import ZEO.ClientStorage
import ZODB.TimeStamp

logging.getLogger().addHandler(logging.StreamHandler())
logging.getLogger().setLevel(logging.WARNING)

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    for a in args:
        a = a.split(':')
        a = a[0], int(a[1])
        print a,
        sys.stdout.flush()
        cs = ZEO.ClientStorage.ClientStorage(a, read_only=True)
        print time.ctime(
            ZODB.TimeStamp.TimeStamp(cs._server.lastTransaction()).timeTime()
            ), '(local)'




##############################################################################
#
# Copyright (c) Zope Corporation.  All Rights Reserved.
#
# This software is subject to the provisions of the Zope Visible Source
# License, Version 1.0 (ZVSL).  A copy of the ZVSL should accompany this
# distribution.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
name = 'zc.zrs'
version = '2.4.2'

from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages

import os, shutil
if os.path.isdir('build'):
    shutil.rmtree('build')

entry_points = """
[console_scripts]
zrsmonitor-script = zc.zrs.monitor:main
last-zeo-transaction = zc.zrs.last:main
"""

tests_require = ['zope.testing', 'zc.zk [static]', 'mock']

long_description = open('README.rst').read()

setup(
    name = name,
    long_description = long_description,
    description = long_description.split('\n')[1],
    version = version,
    author = "Jim Fulton",
    author_email = "jim#zope.com",
    license = "ZPL 2.1",
    keywords = "ZODB",

    packages = ['zc', 'zc.zrs'],
    include_package_data = True,
    data_files = [('.', ['README.txt'])],
    zip_safe = True,
    entry_points = entry_points,
    package_dir = {'':'src'},
    namespace_packages = ['zc'],
    install_requires = [
        'setuptools',
        'ZODB3',
        'Twisted',
        ],
    extras_require = dict(test=tests_require),
    tests_require = tests_require,
    test_suite = 'zc.zrstests.tests.test_suite',
    )



##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Monitor ZRS servers
"""

import ConfigParser
import datetime
import logging
import logging.config
import sys
import threading
import time
import ZEO.ClientStorage, ZEO.Exceptions
import ZODB.TimeStamp
import ZODB.utils

event = threading.Event()
stop = event.set

def main(args=None, testing=False):
    if args is None:
        args = sys.argv[1:]

    [config_path] = args

    logging.config.fileConfig(config_path)
    global logger, monitor_logger
    logger = logging.getLogger(__name__)
    monitor_logger = logging.getLogger('monitor')    
    parser = ConfigParser.RawConfigParser()
    parser.read(config_path)
    data = dict(
        [(section, dict(parser.items(section)))
         for section in parser.sections()]
        )
    options = data["monitor"]
    frequency = float(options.get('frequency', 5))
    message = options.get(
        'message',
        '%(hostname)s %(service)s %(comment)s')

    clusters = [
        Cluster(cluster_name, data[cluster_name], frequency, message)
        for cluster_name in options['clusters'].split()
        ]

    if not testing:
        # This is a round-about way to run forever
        event.wait()
    else:
        return clusters

zeo_options = dict(wait= False, read_only=True,
                   min_disconnect_poll=1, max_disconnect_poll=30,
                   )

class Base(object):

    def __init__(self, name, address_string, service, message):
        address = tuple(address_string.split(':'))
        address = address[0], int(address[1])
        self.address = address
        self.name = name
        self.message = message
        self.service = name + '-' + service
        self.storage = ZEO.ClientStorage.ClientStorage(address, **zeo_options)

    def _report(self, log, severity, comment):
        return log(
            self.message % dict(
                hostname = self.address[0],
                port = self.address[1],
                name = self.name,
                service = self.service,
                severity = severity,
                comment = comment,
                utc = datetime.datetime.utcnow().strftime("%m-%d-%Y %H:%M:%S"),
                ))

    def ok(self, comment):
        self._report(monitor_logger.info, 'INFO', comment)

    def warning(self, comment):
        self._report(monitor_logger.warning, 'WARNING', comment)

    def critical(self, comment):
        self._report(monitor_logger.critical, 'CRITICAL', comment)

class Cluster(Base):

    def __init__(self, name, options, frequency, message):
        super(Cluster, self).__init__(name, options['primary'], 'primary',
                                      message)
        self.frequency = frequency

        self.secondaries = [
            Secondary(name, address_string, message)
            for address_string in options['secondaries'].split()
            ]
        thread = self.thread = threading.Thread(target=self.run, name=name)
        thread.setDaemon(True)
        thread.start()

    def run(self):
        try:
            delta_seconds = self.frequency * 60
            storage = self.storage
            secondaries = self.secondaries
            last_connected = 'Never connected'

            next = time.time()
            next -= next % delta_seconds

            ts_seconds = 0
            
            while not event.isSet():
                now = time.time()
                next += delta_seconds
                wait = next - now
                while wait < 0.0:
                    next += delta_seconds
                    wait = next - now

                time.sleep(wait)
                if event.isSet():
                    break

                try:
                    ts = storage._server.lastTransaction()
                except ZEO.Exceptions.ClientDisconnected:
                    self.critical("Disconnected since: %s" % last_connected)
                else:
                    last_connected = time.ctime(time.time())
                    ts_seconds = ZODB.TimeStamp.TimeStamp(ts).timeTime()
                    if ts == ZODB.utils.z64:
                        self.ok("No transactions")
                    else:
                        self.ok("Committed: %s" % time.ctime(ts_seconds))

                for secondary in secondaries:
                    secondary.check(ts_seconds, now)

        except:
            logger.exception("Error in cluster %s", self.name)
            monitor_logger.error("Error in cluster %s", self.name)
            raise
            
class Secondary(Base):

    def __init__(self, name, address_string, message):
        super(Secondary, self).__init__(name, address_string, 'secondary',
                                        message)
        self.last_connected = 'Never connected'
        self.last_seconds = -2208988800.0

    def check(self, primary_seconds, primary_start):
        try:
            ts = self.storage._server.lastTransaction()
        except ZEO.Exceptions.ClientDisconnected:
            self.critical("Disconnected since: %s" % self.last_connected)
        else:
            self.last_connected = time.ctime(time.time())
            ts_seconds = ZODB.TimeStamp.TimeStamp(ts).timeTime()
            if (ts_seconds - primary_seconds) > (time.time()-primary_start):
                if primary_seconds < 0:
                    self.critical(
                        "Secondary has data, %s, but primary doesn't."
                        % time.ctime(ts_seconds)
                        )
                else:
                    self.critical(
                        "Secondary is ahead of primary s=%s p=%s"
                        % (time.ctime(ts_seconds), time.ctime(primary_seconds))
                        )
            elif ((primary_seconds - ts_seconds) > 60
                  or
                  ((primary_seconds > ts_seconds)
                   and 
                   (time.time() - primary_seconds) > 60
                   )
                ):
                # We're out of date.  Maybe we're making progress.
                if ts_seconds > self.last_seconds:
                    meth = self.warning
                else:
                    meth = self.critical

                if ts == ZODB.utils.z64:
                    meth("Secondary has no data, but primary does: %s" %
                         (time.ctime(primary_seconds))
                         )
                else:
                    meth("Secondary out of date: s=%s p=%s" %
                         (time.ctime(ts_seconds), time.ctime(primary_seconds))
                         )
            else:
                if ts == ZODB.utils.z64:
                    self.ok("No data")
                else:
                    self.ok("Committed: %s" % time.ctime(ts_seconds))

            self.last_seconds = ts_seconds



##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import zc.zrs.primary
import zc.zrs.secondary

class ZRS:

    def __init__(self, config):
        self.config = config
        self.name = config.getSectionName()

    def open(self):
        config = self.config
        replicate_to = config.replicate_to
        replicate_from = config.replicate_from
        zookeeper = config.zookeeper

        storage = config.base.open()

        if replicate_to is not None:
            if replicate_from is not None:
                if replicate_to.address == replicate_from.address:
                    raise ValueError(
                        "Values for replicate-to and "
                        "replicate-from must be different.")

            if zookeeper:
                from zc.zrs.zk import Primary
                storage = Primary(storage, ('', 0),
                                  zookeeper, replicate_to.address)
            else:
                storage = zc.zrs.primary.Primary(storage, replicate_to.address)

        elif replicate_from is None:
            raise ValueError(
                "You must specify replicate-to and/or replicate-from.")

        if replicate_from is not None:
            zookeeper = config.replicate_from_zookeeper or zookeeper
            if zookeeper:
                from zc.zrs.zk import Secondary
                storage = Secondary(
                    storage, zookeeper, replicate_from.address,
                    keep_alive_delay=self.config.keep_alive_delay,
                    )
            else:
                storage = zc.zrs.secondary.Secondary(
                    storage, replicate_from.address,
                    keep_alive_delay=self.config.keep_alive_delay,
                    )

        return storage



##############################################################################
#
# Copyright (c) 2013 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

try:
    from hashlib import md5
except ImportError:
    from md5 import new as md5
from ZODB.TimeStamp import TimeStamp
from zope.testing import setupstack, renormalizing
from zc.zrs import loopback
import doctest
import ZEO.ClientStorage
import ZEO.tests.forker
import ZEO.tests.testZEO
import ZODB.blob
import ZODB.FileStorage
import ZODB.TimeStamp
import ZODB.tests.testblob
import ZODB.tests.testFileStorage
import ZODB.tests.StorageTestBase
import ZODB.utils
import cPickle
import logging
import mock
import os
import re
import shutil
import struct
import subprocess
import sys
import tempfile
import threading
import time
import transaction
import twisted.internet.base
import twisted.internet.error
import twisted.python.failure
import unittest
import zc.zk.testing
import zc.zrs.primary
import zc.zrs.reactor
import zc.zrs.secondary
import zc.zrs.sizedmessage


# start the reactor thread so that it isn't reported as left over:
zc.zrs.reactor.reactor()

def scan_from_back():
    r"""
Create the database:

    >>> from ZODB.DB import DB
    >>> import persistent.dict

    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> db = DB(fs)
    >>> conn = db.open()

    >>> for i in range(100):
    ...     conn.root()[i] = persistent.dict.PersistentDict()
    ...     commit()

Now, be evil, and muck up the beginning: :)

    >>> fs._file.seek(12)
    >>> fs._file.write('\xff'*8)
    >>> conn.root()[100] = persistent.dict.PersistentDict()
    >>> commit()

If we try to iterate from the beginning, we'll get an error:

    >>> condition = threading.Condition()
    >>> it = zc.zrs.primary.FileStorageIterator(fs, condition)
    >>> it.next()
    Traceback (most recent call last):
    ...
    CorruptedDataError: Error reading unknown oid.  Found '' at 4

    >>> def tid_from_time(t):
    ...     return repr(TimeStamp(*(time.gmtime(t)[:5] + (t%60,))))

    >>> tid = tid_from_time(time.time()-70)
    >>> zc.zrs.primary.FileStorageIterator(fs, condition, tid)
    ... # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    OverflowError: ...

But, if we iterate from near the end, we'll be OK:

    >>> tid = tid_from_time(time.time()-30)
    >>> it = zc.zrs.primary.FileStorageIterator(fs, condition, tid)
    >>> trans = it.next()
    >>> from ZODB import utils
    >>> print TimeStamp(trans.tid), [int(utils.u64(r.oid)) for r in trans]
    2007-03-21 20:34:09.000000 [0, 72]

    >>> print TimeStamp(tid)
    2007-03-21 20:34:08.000000

    >>> tid = tid_from_time(time.time()-29.5)
    >>> it = zc.zrs.primary.FileStorageIterator(fs, condition, tid)
    >>> trans = it.next()
    >>> from ZODB import utils
    >>> print TimeStamp(trans.tid), [int(utils.u64(r.oid)) for r in trans]
    2007-03-21 20:34:09.000000 [0, 72]

    >>> print TimeStamp(tid)
    2007-03-21 20:34:08.500000

    """

def primary_suspend_resume():
    """
The primary producer is supposed to be suspendable.

We'll create a file-storage:

    >>> import ZODB.FileStorage
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> from ZODB.DB import DB
    >>> db = DB(fs)

Now, we'll create a special transport that will output data when it is called:

    >>> class Reactor:
    ...     def callFromThread(self, f, *args, **kw):
    ...         f(*args, **kw)

    >>> class Transport:
    ...     def __init__(self):
    ...         self.reactor = Reactor()
    ...     def writeSequence(self, message):
    ...         message = message[1] # cheat. :)
    ...         if message:
    ...             message = cPickle.loads(message)
    ...             if type(message) is tuple:
    ...                 message = message[0]
    ...         print message
    ...     def registerProducer(self, producer, streaming):
    ...         print 'registered producer'
    ...     def unregisterProducer(self):
    ...         print 'unregistered producer'
    ...     def loseConnection(self):
    ...         print 'loseConnection'

And a producer based on the file storage and transport:

    >>> import time
    >>> producer = zc.zrs.primary.PrimaryProducer(
    ...            (fs, None, ZODB.utils.z64), Transport(), 'test'
    ...            ); time.sleep(0.1)
    registered producer
    T
    S
    <class 'persistent.mapping.PersistentMapping'>
    C

We get the initial transaction, because the producer starts producing
immediately.  Let's pause producing:

    >>> producer.pauseProducing()
    DEBUG zc.zrs.primary:
    test pausing

and we'll create another transaction:

    >>> conn = db.open()
    >>> ob = conn.root()
    >>> import persistent.mapping
    >>> ob.x = persistent.mapping.PersistentMapping()
    >>> commit()
    >>> producer.iterator.notify()
    >>> ob = ob.x
    >>> ob.x = persistent.mapping.PersistentMapping()
    >>> commit()
    >>> producer.iterator.notify()
    >>> time.sleep(0.1)

No output because we are paused.  Now let's resume:

    >>> producer.resumeProducing(); time.sleep(0.1)
    DEBUG zc.zrs.primary:
    test resuming
    T
    S
    <class 'persistent.mapping.PersistentMapping'>
    S
    <class 'persistent.mapping.PersistentMapping'>
    C
    T
    S
    <class 'persistent.mapping.PersistentMapping'>
    S
    <class 'persistent.mapping.PersistentMapping'>
    C

and pause again:

    >>> producer.pauseProducing()
    DEBUG zc.zrs.primary:
    test pausing

    >>> ob = ob.x
    >>> ob.x = persistent.mapping.PersistentMapping()
    >>> commit()
    >>> producer.iterator.notify()
    >>> time.sleep(0.1)

and resume:

    >>> producer.resumeProducing(); time.sleep(0.1)
    DEBUG zc.zrs.primary:
    test resuming
    T
    S
    <class 'persistent.mapping.PersistentMapping'>
    S
    <class 'persistent.mapping.PersistentMapping'>
    C

    >>> producer.close(); producer.thread.join()
    unregistered producer
    loseConnection

    >>> db.close()

"""

def secondary_close_edge_cases():
    r"""
There a number of cases to consider when closing a secondary:

- Closing while connecting

  The reactor.clients attribute has a list of secondaries that are
  "connecting". The reactor later attribute has requests to do things
  later.

    >>> reactor.clients
    []

    >>> reactor.later
    []

    >>> import zc.zrs.secondary
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> len(reactor.clients)
    1

    >>> ss.close()
    INFO zc.zrs.secondary:
    Closing Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Stopping factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> reactor.clients
    []

    >>> reactor.later
    []

- Closing while waiting to connect

  We'll reject the connection attempt, which will make the secondary
  queue a connection attempt for later:

    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> reactor.reject()
    INFO zc.zrs.reactor:
    Stopping factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> reactor.clients
    []

    >>> len(reactor.later)
    1

    >>> ss.close()
    INFO zc.zrs.secondary:
    Closing Data.fs ('', 8000)

    >>> reactor.doLater()
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>
    INFO zc.zrs.reactor:
    Stopping factory <zc.zrs.secondary.SecondaryFactory instance at 0xb662b5cc>

    >>> reactor.later
    []

    >>> reactor.clients
    []

- Closing while connected but between transactions

    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor,
    ...         keep_alive_delay=60)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> connection = reactor.accept()
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47248): Connected

    >>> reactor.later
    [<2 60 keep_alive () {}>]

    >>> reactor.clients
    []

    >>> ss.close() # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.secondary:
    Closing Data.fs ('', 8000)
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47248):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>

    >>> reactor.later
    []

    >>> reactor.clients
    []

- Closing while connected and recieving data

    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor,
    ...         keep_alive_delay=60)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> connection = reactor.accept()
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47249): Connected

    >>> connection.read()
    'zrs2.0'
    >>> connection.read()
    '\x00\x00\x00\x00\x00\x00\x00\x00'

    >>> reactor.later
    [<3 60 keep_alive () {}>]

    >>> reactor.clients
    []

    >>> primary_fs = ZODB.FileStorage.FileStorage('primary.fs')
    >>> primary_data = zc.zrs.primary.FileStorageIterator(primary_fs)
    >>> from ZODB.DB import DB
    >>> primary_db = DB(primary_fs)
    >>> trans = primary_data.next()
    >>> connection.send(('T', (trans.tid, trans.status, trans.user,
    ...                        trans.description, trans._extension)))
    >>> record = trans.next
    ... connection.send(('S', (record.oid, record.tid, record.version,
    ...                        record.data, record.data_txn)))

    >>> ss.close() # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.secondary:
    Closing Data.fs ('', 8000)
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47249):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>

    >>> reactor.later
    []

    >>> reactor.clients
    []

    >>> print fs._transaction
    None

    >>> fs._pos
    4L

"""

def primary_data_input_errors():
    r"""
    There is no good reason for a primary to get a data input error. If
    it does, it should simply close the connection.

    >>> import ZODB.FileStorage, zc.zrs.primary
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)

    >>> connection = reactor.connect((('', 8000)))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Connected

    >>> connection.send("Hi") # doctest: +NORMALIZE_WHITESPACE
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Invalid protocol 'Hi'
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    Disconnected <twisted.python.failure.Failure
    <class 'twisted.internet.error.ConnectionDone'>>
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Closed

    >>> connection = reactor.connect((('', 8000)))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246): Connected

    >>> connection.send("xxxxxxxxxxxxxxx") # doctest: +NORMALIZE_WHITESPACE
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246): message too large: (8, 15L)
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246): Closed

    >>> connection = reactor.connect((('', 8000)))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47247): Connected

    >>> connection.send("zrs2.0")
    >>> connection.send("xxxxxxxxxxxxxxx") # doctest: +NORMALIZE_WHITESPACE
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47247): message too large: (8, 15L)
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47247):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47247): Closed

    >>> connection = reactor.connect((('', 8000)))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47248): Connected

    >>> connection.send("zrs2.0")
    >>> connection.send("xxxxxxx") # doctest: +NORMALIZE_WHITESPACE
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47248): Invalid transaction id, 'xxxxxxx'
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47248):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47248): Closed

    Sending any message other than an empty message to the primary
    after the first two messages will result in an error:

    >>> connection = reactor.connect((('', 8000)))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47249): Connected

    >>> connection.send("zrs2.0")
    >>> connection.send("\0"*8) # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47249):
    start '\x00\x00\x00\x00\x00\x00\x00\x00' (1900-01-01 00:00:00.000000)

    >>> connection.send("")
    DEBUG zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47249): keep-alive
    >>> connection.send("")
    DEBUG zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47249): keep-alive

    >>> connection.send("Hi") # doctest: +ELLIPSIS
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47249): Too many messages
    ...
    IPv4Address(TCP, '127.0.0.1', 47249): Closed

    """

def secondary_data_input_errors():
    r"""

There is not good reason for a secondary to get a data input error. If
it does, it should simply close.

    >>> import zc.zrs.secondary
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> connection = reactor.accept()
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47245): Connected

    >>> connection.read()
    'zrs2.0'
    >>> connection.read()
    '\x00\x00\x00\x00\x00\x00\x00\x00'

    >>> connection.send('hi', raw=True)
    ... # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    CRITICAL zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47245): Input data error
    Traceback (most recent call last):
    ...
    BadPickleGet: 105
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    Disconnected <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>

    >>> reactor.later
    []

    >>> reactor.clients
    []

    """


def crashing_reactor_logs_as_such():
    """

We'll write a silly script that simply starts the reactor and tells it
to crash:

    >>> open('t.py', 'w').write('''
    ... import logging, time
    ... import twisted.internet
    ... import zc.zrs.reactor
    ...
    ... logging.getLogger().setLevel(1)
    ... handler = logging.StreamHandler(open('t.log', 'w'))
    ... logging.getLogger().addHandler(handler)
    ... zc.zrs.reactor.reactor()
    ... time.sleep(0.1)
    ... twisted.internet.reactor.callFromThread(twisted.internet.reactor.crash)
    ... time.sleep(0.1)
    ... #logging.error('failed')
    ... ''')

We'll run it:

    >>> env = os.environ.copy()
    >>> env['PYTHONPATH'] = os.pathsep.join(sys.path)
    >>> p = subprocess.Popen(
    ...       [sys.executable, 't.py'],
    ...       env=env)

It exits with a non-zero exit status:

    >>> bool(p.wait())
    False

And we get something in the log to the effect that it closed unexpectedly.

    >>> print open('t.log').read(),
    Main loop terminated.
    The twisted reactor quit unexpectedly

OTOH, if we exit without crashing:

    >>> open('t.py', 'w').write('''
    ... import logging, time
    ... import twisted.internet
    ... import zc.zrs.reactor
    ...
    ... logging.getLogger().setLevel(1)
    ... handler = logging.StreamHandler(open('t.log', 'w'))
    ... logging.getLogger().addHandler(handler)
    ... zc.zrs.reactor.reactor()
    ... time.sleep(0.1)
    ... ''')

    >>> p = subprocess.Popen(
    ...       [sys.executable, 't.py'],
    ...       env=env)

    >>> bool(p.wait())
    False

    >>> print open('t.log').read(),
    Main loop terminated.

    """

def leaking_file_handles_when_secondaries_disconnect():
    r"""

    >>> import sys, ZODB.FileStorage, zc.zrs.primary
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)

    >>> oldrc = sys.getrefcount(zc.zrs.primary.FileStorageIterator)

    >>> from ZODB.DB import DB
    >>> import persistent.dict
    >>> db = DB(ps)
    >>> conn = db.open()
    >>> ob = conn.root()
    >>> for i in range(10):
    ...   ob.x = persistent.dict.PersistentDict()
    ...   commit()

    >>> for i in range(10):
    ...     connection = reactor.connect(('', 8000))
    ...     connection.send("zrs2.0")
    ...     connection.send("\0"*8)
    ...     _ = connection.read()
    ...     connection.close()
    ...     # doctest: +ELLIPSIS
    INFO ...

    >>> time.sleep(.01)
    >>> sys.getrefcount(zc.zrs.primary.FileStorageIterator) == oldrc
    True

    >>> db.close() # doctest: +ELLIPSIS
    INFO ...

    """

def close_writes_new_transactions():
    r"""
    We want close to try to write pending transactions, even if it
    means that close will take a long time.

    >>> import ZODB.FileStorage, zc.zrs.primary, persistent.dict
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)

    >>> db = ZODB.DB(ps)
    >>> conn = db.open()
    >>> ob = conn.root()
    >>> ob.x = 0
    >>> commit()

    >>> committed = 2

    We'll open lots of connections: :)

    >>> nconnections = 10
    >>> connections = []
    >>> for i in range(nconnections):
    ...     connection = reactor.connect(('', 8000))
    ...     connection.send("zrs2.0")
    ...     connection.send("\0"*8)
    ...     connections.append(connection)
    ...     # doctest: +ELLIPSIS
    INFO zc.zrs.primary:...

    >>> import time
    >>> time.sleep(.1)

    >>> for i in range(300):
    ...     ob[i] = persistent.dict.PersistentDict()

    >>> commit()
    >>> committed += 1

    >>> db.close()
    ...     # doctest: +ELLIPSIS
    INFO zc.zrs.primary:...

    >>> for i in range(nconnections):
    ...     connection = connections[i]
    ...     trans = message_type = x = None
    ...     ntrans = 0
    ...     while connection.have_data():
    ...         message_type, data = connection.read()
    ...         if message_type == 'T':
    ...             trans = data
    ...             ntrans += 1
    ...         elif message_type == 'S':
    ...             x = connection.read(True)
    ...     if message_type != 'C' or ntrans != committed:
    ...         print i, message_type, ntrans

    """

def secondary_gives_a_tid_that_is_too_high():
    r"""
    We should error and close the connection if a secondary presents a
    tid that is higher than the largest tid seen by the primary.

    >>> import ZODB.FileStorage
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> import zc.zrs.primary
    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)

    >>> from ZODB.DB import DB
    >>> import persistent.dict
    >>> db = DB(ps)
    >>> conn = db.open()
    >>> ob = conn.root()
    >>> ob.x = persistent.dict.PersistentDict()
    >>> commit()

    >>> import ZODB.utils
    >>> too_high_tid = ZODB.utils.p64(ZODB.utils.u64(ob._p_serial)+1)

    >>> connection = reactor.connect(('', 8000))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Connected

    >>> connection.send("zrs2.0") # doctest: +NORMALIZE_WHITESPACE
    >>> connection.send(too_high_tid); time.sleep(.01) # wait for thread :(
    ... # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    start '\x03lk\x90\xf7wwx' (2007-03-21 20:32:58.000000)
    ERROR zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    Traceback (most recent call last):
    ...
    TidTooHigh: '\x03lk\x90\xf7wwx'
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    Disconnected
    <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>

    """


class FauxScanControl(object):
    _v = True
    _n = 0
    def get(self):
        self._n += 1
        if self._n > 10:
            time.sleep(.01)
        if self._n > 20:
           print self._n
        return self._v
    def set(self, v):
        self._v = v
    not_stopped = property(get, set)

def scan_control_stops_scans_on_client_disconnects():
    r"""
    We want to be able to limit iterator scans if a client disconnects, for
    example, to limit impact on the server if a large scan is required.

    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> import persistent.dict
    >>> db = ZODB.DB(fs)
    >>> conn = db.open()
    >>> ob = conn.root()
    >>> for i in range(100):
    ...     ob[i] = persistent.dict.PersistentDict()
    ...     commit()
    >>> tid1 = ob._p_serial
    >>> for i in range(100, 200):
    ...     ob[i] = persistent.dict.PersistentDict()
    ...     commit()
    >>> tid2 = ob._p_serial
    >>> for i in range(200, 300):
    ...     ob[i] = persistent.dict.PersistentDict()
    ...     commit()

    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)

    >>> import time
    >>> ScanControl = zc.zrs.primary.ScanControl
    >>> zc.zrs.primary.ScanControl = FauxScanControl

    >>> connection = reactor.connect(('', 8000))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Connected

    >>> connection.send("zrs2.0") # doctest: +NORMALIZE_WHITESPACE
    >>> connection.send(tid1) # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245):
    start '\x03lk\x92\x9d\xdd\xdd\xdd' (2007-03-21 20:34:37.000000)
    >>> connection.loseConnection() # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47245): Disconnected
    <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>


    >>> connection = reactor.connect(('', 8000))
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246): Connected

    >>> connection.send("zrs2.0") # doctest: +NORMALIZE_WHITESPACE
    >>> connection.send(tid2) # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246):
    start '\x03lk\x94H\x88\x88\x88' (2007-03-21 20:36:17.000000)

    >>> connection.loseConnection() # doctest: +NORMALIZE_WHITESPACE
    INFO zc.zrs.primary:
    IPv4Address(TCP, '127.0.0.1', 47246): Disconnected
    <twisted.python.failure.Failure
      <class 'twisted.internet.error.ConnectionDone'>>

    >>> time.sleep(.1)
    >>> zc.zrs.primary.ScanControl = ScanControl
    """

def record_iternext():
    """
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')
    >>> ps = zc.zrs.primary.Primary(fs, ('', 8000), reactor)
    INFO zc.zrs.primary:
    Opening Data.fs ('', 8000)
    >>> ss = zc.zrs.secondary.Secondary(ps, ('', 9000), reactor)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 9000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance at 0xb654796c>

    >>> ps.record_iternext == fs.record_iternext
    True

    >>> ss.record_iternext == fs.record_iternext
    True

    """

def is_blob_record():
    r"""
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs', blob_dir='blobs')
    >>> db = ZODB.DB(fs)
    >>> conn = db.open()
    >>> conn.root()['blob'] = ZODB.blob.Blob()
    >>> transaction.commit()
    >>> zc.zrs.primary.is_blob_record(fs.load(ZODB.utils.p64(0), '')[0])
    False
    >>> zc.zrs.primary.is_blob_record(fs.load(ZODB.utils.p64(1), '')[0])
    True

    An invalid pickle yields a false value:

    >>> zc.zrs.primary.is_blob_record("Hello world!")
    False
    >>> zc.zrs.primary.is_blob_record('c__main__\nC\nq\x01.')
    False

    >>> db.close()
    """

def secondary_replicate_from_old_zrs_that_doesnt_send_checksums():
    r"""
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')

    >>> import zc.zrs.secondary
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor,
    ...                                 check_checksums=False)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> connection = reactor.accept()
    INFO zc.zrs.secondary:
    IPv4Address(TCP, '127.0.0.1', 47245): Connected

    >>> connection.read()
    'zrs2.0'

    >>> connection.read()
    '\x00\x00\x00\x00\x00\x00\x00\x00'

    >>> primary_fs = ZODB.FileStorage.FileStorage('primary.fs')
    >>> import zc.zrs.primary
    >>> primary_data = zc.zrs.primary.FileStorageIterator(primary_fs)

    >>> from ZODB.DB import DB
    >>> primary_db = DB(primary_fs)

    >>> trans = primary_data.next()

    >>> connection.send(('T', (trans.tid, trans.status, trans.user,
    ...                        trans.description, trans._extension)))

    >>> for record in trans:
    ...     connection.send(('S', (record.oid, record.tid, record.version,
    ...                            record.data_txn)))
    ...     connection.send(record.data, raw=True)

    >>> connection.send(('C', ()))

    >>> _ = fs.load('\0'*8, '')

    """

def secondary_gets_extension_data():
    """
    >>> fs = ZODB.FileStorage.FileStorage('Data.fs')

    >>> import zc.zrs.secondary
    >>> ss = zc.zrs.secondary.Secondary(fs, ('', 8000), reactor,
    ...                                 check_checksums=False)
    INFO zc.zrs.secondary:
    Opening Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Starting factory <zc.zrs.secondary.SecondaryFactory instance>

    >>> for n in ss.getExtensionMethods():
    ...     if getattr(ss, n) != getattr(fs, n):
    ...         print 'oops'

    >>> ss.close()
    INFO zc.zrs.secondary:
    Closing Data.fs ('', 8000)
    INFO zc.zrs.reactor:
    Stopping factory <zc.zrs.secondary.SecondaryFactory instance at 0x2bebb00>
    """

class DelayedCall:

    def __init__(self, later, n, delay, func, args, kw):
        self.later, self.n, self.delay = later, n, delay
        self.func, self.args, self.kw = func, args, kw

    def __repr__(self):
        return "<%s %s %s %r %r>" % (
            self.n, self.delay, self.func.__name__, self.args, self.kw)

    def cancel(self):
        if self in self.later:
            self.later.remove(self)
        self.later = None

    def active(self):
        return self.later is not None

    def __call__(self):
        self.later = None
        self.func(*self.args, **self.kw)

class TestReactor:

    def __init__(self):
        self._factories = {}
        self.clients = []
        self.client_port = 47245
        self.later = []

    def listenTCP(self, port, factory, backlog=50, interface=''):
        addr = interface, port
        assert addr not in self._factories
        self._factories[addr] = factory
        return TestListener(self, addr)

    def connect(self, addr):
        proto = self._factories[addr].buildProtocol(addr)
        transport = PrimaryTransport(self, addr, self.client_port, proto)
        self.client_port += 1
        proto.makeConnection(transport)
        return transport

    lock = threading.RLock()
    def callFromThread(self, f, *a, **k):
        self.lock.acquire()
        try:
            f(*a, **k)
        finally:
            self.lock.release()

    callLater_n = 0
    def callLater(self, delay, f, *a, **k):
        self.callLater_n += 1
        f = DelayedCall(self.later, self.callLater_n, delay, f, a, k)
        self.later.append(f)
        return f

    def doLater(self):
        l = len(self.later)
        later = self.later[:l]
        del self.later[:l]
        while later:
            self.callFromThread(later.pop(0))

    def connectTCP(self, host, port, factory, timeout=30):
        addr = host, port
        connector = TestConnector(self, addr, factory)
        connector.connect()
        return connector

    def accept(self):
        connector = self.clients.pop(0)
        return connector.accept()

    def reject(self):
        connector = self.clients.pop(0)
        return connector.reject()

class TestListener:

    def __init__(self, reactor, addr):
        self.reactor = reactor
        self.addr = addr

    def stopListening(self):
        if self.addr in self.reactor._factories:
            del self.reactor._factories[self.addr]

close_reason = twisted.python.failure.Failure(
    twisted.internet.error.ConnectionDone())

class MessageTransport:

    def __init__(self, reactor, addr, port, proto=None):
        self.data = ''
        self.cond = threading.Condition()
        self.closed = False
        self.reactor = reactor
        self.addr = addr
        self.peer = "IPv4Address(TCP, '127.0.0.1', %s)" % port
        self.proto = proto
        self.init_md5('\x00\x00\x00\x00\x00\x00\x00\x00')

    def getPeer(self):
        return self.peer

    def write(self, data):
        self.cond.acquire()
        self.data += data
        self.cond.notifyAll()
        self.cond.release()

    def writeSequence(self, data):
        self.write(''.join(data))

    def read(self):
        self.cond.acquire()

        if len(self.data) < 4:
            self.cond.wait(5)
            assert len(self.data) >= 4
        l, = struct.unpack(">I", self.data[:4])
        self.data = self.data[4:]

        if len(self.data) < l:
            self.cond.wait(5)
            assert len(self.data) >= l, (l, len(self.data))
        result = self.data[:l]
        self.data = self.data[l:]

        self.cond.release()

        return result

    def init_md5(self, data):
        self.md5 = md5(data)

    def send(self, data):
        record = zc.zrs.sizedmessage.marshal(data)
        self.md5.update(record[4:])
        dataReceived = self.proto.dataReceived

        # send data in parts to try to confuse the protocol
        n = 1
        while record:
            data, record = record[:n], record[n:]
            if data and not self.closed:
                dataReceived(data)
            n *= 2

    def have_data(self):
        return bool(self.data)

    def loseConnection(self):
        self.closed = True
        self.proto.connectionLost(close_reason)

    producer = None
    def registerProducer(self, producer, streaming):
        self.producer = producer

    def unregisterProducer(self):
        if self.producer is not None:
            self.producer = None
            if self.closed:
                self.proto.connectionLost(close_reason)

    def close(self):
        self.producer.stopProducing()
        self.proto.connectionLost(close_reason)

class PrimaryTransport(MessageTransport):

    def read(self, raw=False):
        data = MessageTransport.read(self)
        if raw:
            return data
        return cPickle.loads(data)

class SecondaryTransport(MessageTransport):

    def send(self, data, raw=False):
        if not raw:
            data = cPickle.dumps(data)
        MessageTransport.send(self, data or '')

    def fail(self):
        self.connectionLost('failed')

    def connectionLost(self, reason):
        """Notify of a lost connection

        This is a distillation of what happens in a tcp Client transport.
        """
        self.proto.connectionLost(reason)
        self.connector.connectionLost(reason)


    def failIfNotConnected(self, reason):
        if self.connector in self.reactor.clients:
            self.reactor.clients.remove(self.connector)
        self.connector.connectionFailed(reason)

class TestConnector(twisted.internet.base.BaseConnector):

    def __init__(self, reactor, addr, factory, timeout=None):
        twisted.internet.base.BaseConnector.__init__(
            self, factory, timeout, reactor)
        self.addr = addr

    def _makeTransport(self):
        reactor, addr = self.reactor, self.addr

        if addr in reactor._factories:
            # We have a server and a client.  We'll hook them together via
            # a loopback mechanism
            proto = self.buildProtocol(addr)
            server = reactor._factories[addr].buildProtocol(addr)
            loopback.loopbackAsync(server, proto, self)
            transport = proto.transport
        else:
            reactor.clients.append(self)
            transport = SecondaryTransport(reactor, addr, reactor.client_port)
            reactor.client_port += 1

        transport.connector = self
        return transport

    def accept(self):
        reactor, addr, transport = self.reactor, self.addr, self.transport
        proto = self.buildProtocol(addr)
        transport.proto = proto
        transport.connector = self
        proto.makeConnection(transport)
        return transport

    def reject(self):
        self.connectionFailed('rejected')


class Stdout:
    def write(self, data):
        sys.stdout.write(data)
    def flush(self):
        sys.stdout.flush()

stdout_handler = logging.StreamHandler(Stdout())
stdout_handler.setFormatter(logging.Formatter(
    "%(levelname)s %(name)s:\n%(message)s"))

def join(old):
    # Wait for any new threads created during a test to die.
    for thread in threading.enumerate():
        if thread not in old:
            thread.join(1.0)

def setUpTime(test):
    setupstack.register(test, join, threading.enumerate())
    setupstack.setUpDirectory(test)
    global now
    now = time.mktime((2007, 3, 21, 15, 32, 57, 2, 80, 0))
    oldtime = time.time
    setupstack.register(test, lambda : setattr(time, 'time', oldtime))
    time.time = lambda : now
    def commit():
        global now
        now += 1
        transaction.commit()
    test.globs['commit'] = commit

def setUp(test):
    setUpTime(test)

    test.globs['reactor'] = TestReactor()

    logger = logging.getLogger('zc.zrs')
    logger.setLevel(1)
    setupstack.register(test, logger.setLevel, 0)
    logger.addHandler(stdout_handler)
    setupstack.register(test, logger.removeHandler, stdout_handler)

##############################################################################
# Reuse ZODB Storage Tests

from ZODB.tests import StorageTestBase
from ZODB.tests import BasicStorage
from ZODB.tests import TransactionalUndoStorage
from ZODB.tests import RevisionStorage
from ZODB.tests import PackableStorage
from ZODB.tests import Synchronization
from ZODB.tests import ConflictResolution
from ZODB.tests import HistoryStorage
from ZODB.tests import IteratorStorage
from ZODB.tests import PersistentStorage
from ZODB.tests import MTStorage
from ZODB.tests import ReadOnlyStorage

class TestPrimary(zc.zrs.primary.Primary):

    _transaction_count = 0
    def tpc_finish(self, *args):
        self._transaction_count += 1
        zc.zrs.primary.Primary.tpc_finish(self, *args)
        if self._transaction_count%20 == 0:
            # be annoying and disconnect our clients every 20 transactions.
            # Hee hee.
            # Before we do that though, we'll call doLater on our reactor to
            # give previously disconnected clients a chance to reconnect.
            self._reactor.doLater()
            for instance in self._factory.instances:
                instance._stop()


class BasePrimaryStorageTests(StorageTestBase.StorageTestBase):

    use_blob_storage = False

    def setUp(self):
        self.__pack = None

        self.globs = {}
        setupstack.register(self, join, threading.enumerate())
        setupstack.setUpDirectory(self)
        self.globs['reactor'] = TestReactor()
        self.open(create=1)

    def tearDown(self):
        # Give any disconnected clients a chance to reconnect.
        self._storage._reactor.doLater()

        self._storage.close()
        reactor = self.globs['reactor']
        self.assert_(not reactor._factories) # Make sure we're not listening
        setupstack.tearDown(self)
        self.globs.clear()

    __port = 8000

    def catch_up(self, fs1, fs2):
        for i in range(2000):
            self._storage._reactor.doLater()
            if i:
                time.sleep(0.01)

            if fs1.lastTransaction() <= fs2.lastTransaction():
                return # caught up

            l1 = list(fs1.iterator())
            if not l1:
                return
            l2 = list(fs2.iterator())
            if l2:
                if (l1[-1].tid <= l2[-1].tid):
                    return

        raise AssertionError("Can't catch up.")


    def open(self, **kwargs):
        reactor = self.globs['reactor']
        self.__port += 1
        addr = '', self.__port
        if self.use_blob_storage:
            self.__pfs = ZODB.FileStorage.FileStorage(
                'primary.fs', blob_dir='primary_blobs', **kwargs)
            self.__sfs = ZODB.FileStorage.FileStorage(
                'secondary.fs', blob_dir='secondary_blobs')
        else:
            self.__pfs = ZODB.FileStorage.FileStorage('primary.fs', **kwargs)
            self.__sfs = ZODB.FileStorage.FileStorage('secondary.fs')

        self._storage = self._wrap(TestPrimary(self.__pfs, addr, reactor))
        self.__ss = zc.zrs.secondary.Secondary(
            self._wrap(self.__sfs), addr, reactor)

        p_pack = self._storage.pack
        def pack(*args, **kw):
            p_pack(*args, **kw)
            self.__pack = True
        self._storage.pack = pack

        p_close = self._storage.close
        def close():
            self.catch_up(self.__pfs, self.__sfs)

            if self.__pack:
                comparedbs_packed(self, self.__pfs, self.__sfs)
            else:
                self.__comparedbs(self.__pfs, self.__sfs)

            # Now, just recover from scratch to make sure we can:
            if self.use_blob_storage:
                sfs = ZODB.FileStorage.FileStorage(
                    'secondary2.fs', blob_dir='secondarys_blobs')
            else:
                sfs = ZODB.FileStorage.FileStorage('secondary2.fs')

            ss = zc.zrs.secondary.Secondary(sfs, addr, reactor)
            self.catch_up(self.__pfs, sfs)
            self.__comparedbs(self.__pfs, sfs)
            ss.close()

            p_close()
            self.__ss.close()

        self._storage.close = close

    _wrap = lambda self, s: s

    def __comparedbs(self, fs1, fs2):
        if fs1._pos != fs2._pos:
            time.sleep(0.1)
        self.assertEqual(fs1._pos, fs2._pos)

        self.compare(fs1, fs2)



def comparedbs_packed(self, fs1, fs2):

    # The primary was packed.  This introduces some significant
    # complications.  The secondary can end up with packed records
    # following unpacked records, depending on timing.  Or, it can
    # end up with packed records that don't exist on the primary,
    # except in pathological cases.  This can lead to the
    # secondary having extra records not present on the
    # primary. In any case it isn't reasonable to expect the
    # secondary to have exactly the same records as the primary.
    # We'll do a looser comparison that requires:
    #
    # - The primary's records are a subset of the secondary's, and
    # - The primary and secondary have the same current records.
    time.sleep(0.01)

    data2 = {}
    current2 = {}
    for trans in fs2.iterator():
        objects = {}
        data2[trans.tid] = (trans.user, trans.description,
                            trans._extension, objects)
        for r in trans:
            objects[r.oid] = r.tid, r.version, r.data
            current2[r.oid] = trans.tid


    current1 = {}
    for trans in fs1.iterator():
        self.assertEqual(
            data2[trans.tid][:3],
            (trans.user, trans.description, trans._extension),
            )
        objects = {}
        for r in trans:
            objects[r.oid] = r.tid, r.version, r.data
            current2[r.oid] = trans.tid

        objects2 = data2[trans.tid][3]
        for oid in objects:
            self.assertEqual(objects2[r.oid], objects[r.oid])

    for oid, tid in current1.items():
        self.assertEqual(current2[oid], tid)

def tsr(tid):
    return repr(str(TimeStamp(tid)))

def show_fs(fs):
    for t in fs.iterator():
        print tsr(t.tid), repr(t.status), repr(t.description), t._pos
        for r in t:
            print ' ', ZODB.utils.u64(r.oid), tsr(r.tid), repr(r.version),
            print r.data and len(r.data), r.data_txn and tsr(r.data_txn), r.pos

class PrimaryStorageTests(
    BasePrimaryStorageTests,
    BasicStorage.BasicStorage,
    TransactionalUndoStorage.TransactionalUndoStorage,
    RevisionStorage.RevisionStorage,
    PackableStorage.PackableStorage,
    PackableStorage.PackableUndoStorage,
    Synchronization.SynchronizedStorage,
    ConflictResolution.ConflictResolvingStorage,
    ConflictResolution.ConflictResolvingTransUndoStorage,
    HistoryStorage.HistoryStorage,
    IteratorStorage.IteratorStorage,
    IteratorStorage.IteratorDeepCompare,
    IteratorStorage.ExtendedIteratorStorage,
    PersistentStorage.PersistentStorage,
    MTStorage.MTStorage,
    ReadOnlyStorage.ReadOnlyStorage
    ):
    pass

class PrimaryStorageTestsWithBobs(PrimaryStorageTests):

    use_blob_storage = True

class PrimaryHexStorageTestsWithBobs(PrimaryStorageTestsWithBobs):

    def _wrap(self, s):
        return zc.zrs.xformstorage.HexStorage(s)

#
##############################################################################

##############################################################################
# ZEO Tests

class ZEOTests(ZEO.tests.testZEO.FullGenericTests):

    def getConfig(self):
        port = self._ZEOTests_port = ZEO.tests.testZEO.get_port()
        return """
        %%import zc.zrs

        <zrs>
          replicate-to %s
          <filestorage 1>
            path primary.fs
          </filestorage>
        </zrs>
        """ % port

    def setUp(self):
        self.globs = {}
        setupstack.register(self, join, threading.enumerate())
        setupstack.setUpDirectory(self)
        ZEO.tests.testZEO.FullGenericTests.setUp(self)
        self.__sfs = ZODB.FileStorage.FileStorage('secondary.fs')
        self.__s = zc.zrs.secondary.Secondary(
            self._wrap(self.__sfs),
            ('', self._ZEOTests_port), reconnect_delay=0.1)
        zc.zrs.reactor.reactor().callLater(0.1, self.__breakConnection)

    def __breakConnection(self):
        try:
            f = self.__s._factory.instance.transport.loseConnection
        except AttributeError:
            return
        f('broken by test')

    def tearDown(self):

        # Check whether secondary has same data as primary:
        for i in range(1000):
            try:
                fsp = ZODB.FileStorage.FileStorage('primary.fs',
                                                   read_only=True)
                comparedbs_packed(self, fsp, self.__sfs)
                break
            except:
                # Hm. Maybe we didn't wait long enough before starting
                # the compare.  Let's wait a tad longer.
                if i == 999:
                    raise
                time.sleep(.1)

        fsp.close()
        self.__s.close()
        ZEO.tests.testZEO.FullGenericTests.tearDown(self)
        setupstack.tearDown(self)

    _wrap = lambda self, s: s


class BlobWritableCacheTests(ZEO.tests.testZEO.BlobWritableCacheTests):

    def getConfig(self):
        port = self._ZEOTests_port = ZEO.tests.testZEO.get_port()
        return """
        %%import zc.zrs

        <zrs>
          replicate-to %s
          <filestorage 1>
            blob-dir blobs
            path primary.fs
          </filestorage>
        </zrs>
        """ % port

class ZEOHexTests(ZEOTests):

    def getConfig(self):
        port = self._ZEOTests_port = ZEO.tests.testZEO.get_port()
        return """
        %%import zc.zrs
        %%import zc.zrs.xformstorage

        <hexstorage>
          <zrs>
            replicate-to %s
            <filestorage 1>
              path primary.fs
            </filestorage>
          </zrs>
        </hexstorage>
        """ % port

    def _wrap(self, s):
        return zc.zrs.xformstorage.HexStorage(s)

class ZEOHexClientHexTests(ZEOHexTests):

    def _wrap_client(self, s):
        return zc.zrs.xformstorage.HexStorage(s)

class ZEOHexClientTests(ZEOHexTests):

    def getConfig(self):
        port = self._ZEOTests_port = ZEO.tests.testZEO.get_port()
        return """
        %%import zc.zrs
        %%import zc.zrs.xformstorage

        <serverhexstorage>
          <zrs>
            replicate-to %s
            <filestorage 1>
              path primary.fs
            </filestorage>
          </zrs>
        </serverhexstorage>
        """ % port

    def _wrap(self, s):
        return zc.zrs.xformstorage.HexStorage(s, True)

    def _wrap_client(self, s):
        return zc.zrs.xformstorage.HexStorage(s)


#
##############################################################################

##############################################################################
# Monitor test

def monitor_setUp(test):
    test.globs['__logging_dict'] = dict(logging.__dict__)
    reload(logging)

def monitor_tearDown(test):
    logging.__dict__.clear()
    logging.__dict__.update(test.globs['__logging_dict'])

#
##############################################################################

##############################################################################
# Run ZODB hex tests with our hex storage. We're doing this mainly
# to test our hex storage so we can use it to test zrs. :)
# The reason we need our own hex storage is that we need one
# that won't choke if records are already hexed to test some of the scenarios
# we care about.

class FileStorageHexTests(ZODB.tests.testFileStorage.FileStorageTests):
    def open(self, **kwargs):
        self._storage = zc.zrs.xformstorage.HexStorage(
            ZODB.FileStorage.FileStorage('FileStorageTests.fs',**kwargs))

class FileStorageHexTestsWithBlobsEnabled(
    ZODB.tests.testFileStorage.FileStorageTests
    ):
    def open(self, **kwargs):
        if 'blob_dir' not in kwargs:
            kwargs = kwargs.copy()
            kwargs['blob_dir'] = 'blobs'
        ZODB.tests.testFileStorage.FileStorageTests.open(self, **kwargs)
        self._storage = zc.zrs.xformstorage.HexStorage(self._storage)

class FileStorageHexRecoveryTest(
    ZODB.tests.testFileStorage.FileStorageRecoveryTest):
    def setUp(self):
        ZODB.tests.StorageTestBase.StorageTestBase.setUp(self)
        self._storage = zc.zrs.xformstorage.HexStorage(
            ZODB.FileStorage.FileStorage("Source.fs", create=True))
        self._dst = zc.zrs.xformstorage.HexStorage(
            ZODB.FileStorage.FileStorage("Dest.fs", create=True))

class FileStorageHexTests(ZEO.tests.testZEO.FileStorageTests):
    _expected_interfaces = (
        ('ZODB.interfaces', 'IStorageRestoreable'),
        ('ZODB.interfaces', 'IStorageIteration'),
        ('ZODB.interfaces', 'IStorageUndoable'),
        ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
        ('ZODB.interfaces', 'IExternalGC'),
        ('ZODB.interfaces', 'IStorage'),
        ('ZODB.interfaces', 'IStorageWrapper'),
        ('zope.interface', 'Interface'),
        )

    def getConfig(self):
        return """\
        %import zc.zrs.xformstorage
        <hexstorage>
        <filestorage 1>
        path Data.fs
        </filestorage>
        </hexstorage>
        """

class FileStorageClientHexTests(FileStorageHexTests):

    def getConfig(self):
        return """\
        %import zc.zrs.xformstorage
        <serverhexstorage>
        <filestorage 1>
        path Data.fs
        </filestorage>
        </serverhexstorage>
        """

    def _wrap_client(self, client):
        return ZODB.tests.hexstorage.HexStorage(client)

#
##############################################################################

def setUpZK(test):
    setupstack.setUpDirectory(test)
    setupstack.context_manager(
        test, mock.patch('socket.getfqdn')).return_value = 'localhost'
    zc.zk.testing.setUp(test)

def setUpZKConfig(test):
    setupstack.context_manager(test, mock.patch('ZODB.FileStorage.FileStorage'))
    setupstack.context_manager(test, mock.patch('zc.zrs.zk.Primary'))
    setupstack.context_manager(test, mock.patch('zc.zrs.zk.Secondary'))


def test_suite():
    suite = unittest.TestSuite((
        doctest.DocFileSuite(
            'fsiterator.test',
            'primary.test', 'primary-blob.test', 'primary-blobstorage.test',
            'secondary.test', 'secondary-blob.test',
            'secondary-blobstorage.test',
            setUp=setUp, tearDown=setupstack.tearDown,
            checker=renormalizing.RENormalizing([
                (re.compile(' at 0x[a-fA-F0-9]+'), ''),
                ]),
            ),
        doctest.DocFileSuite(
            'zk.test',
            setUp=setUpZK, tearDown=setupstack.tearDown,
            checker=renormalizing.RENormalizing([
                (re.compile(r"PrimaryFactory starting on \d+"),
                 "PrimaryFactory starting on EPORT"),
                (re.compile(' at 0x[a-fA-F0-9]+'), ''),
                (re.compile(r"/127.0.0.1:\d+"), "/127.0.0.1:PORT"),
                (re.compile(r"/localhost:\d+"), "/127.0.0.1:PORT"),
                (re.compile(r"'127.0.0.1', \d+"), "'127.0.0.1', PORT"),
                (re.compile(r"pid = \d+"), "pid = PORT"),
                ]),
            ),
        doctest.DocFileSuite(
            'zkconfig.test',
            setUp=setUpZKConfig, tearDown=setupstack.tearDown,
            ),
        doctest.DocFileSuite(
            'config.test',
            checker=renormalizing.RENormalizing([
                (re.compile(' at 0x[a-fA-F0-9]+'), ''),
                ]),
            setUp=setUpTime, tearDown=setupstack.tearDown,
            ),
        doctest.DocTestSuite(
            setUp=setUp, tearDown=setupstack.tearDown,
            checker=renormalizing.RENormalizing([
                (re.compile(' at 0x[a-fA-F0-9]+'), ''),
                ]),
            ),
        ))

    def make(class_, *args):
        s = unittest.makeSuite(class_, *args)
        s.layer = ZODB.tests.util.MininalTestLayer(class_.__name__)
        suite.addTest(s)

    make(PrimaryStorageTests, "check")
    make(PrimaryStorageTestsWithBobs, "check")
    make(ZEOTests, "check")
    make(BlobWritableCacheTests, "check")

    try:
        from ZODB.tests import hexstorage
        import zc.zrs.xformstorage

        make(FileStorageHexTests, "check")
        make(FileStorageHexTestsWithBlobsEnabled, "check")
        make(FileStorageHexRecoveryTest, "check")
        make(ZEOHexTests, "check")
        make(ZEOHexClientHexTests, "check")
        make(ZEOHexClientTests, "check")

        suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
            'BlobFileHexStorage',
            lambda name, blob_dir:
            zc.zrs.xformstorage.HexStorage(
                ZODB.FileStorage.FileStorage(
                    '%s.fs' % name, blob_dir=blob_dir)),
            test_blob_storage_recovery=True,
            test_packing=True,
            ))
    except ImportError:
        pass # pre wrapper-support ZODB

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')

